# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Job Typing"""

import json
from abc import ABC
from dataclasses import asdict, dataclass, field
from typing import Any, List

from qibo.models.circuit import Circuit  # type: ignore[import-untyped]
from typeguard import typechecked

from qiboconnection.typings.enums import JobStatus, JobType
from qiboconnection.typings.requests import JobRequest
from qiboconnection.typings.responses.job_response import JobResponse
from qiboconnection.typings.vqa import VQA
from qiboconnection.util import compress_any

from .algorithm import ProgramDefinition
from .devices.device import Device
from .job_result import JobResult
from .user import User


@dataclass
class Job(ABC):
    """Job class to manage the job experiment to be remotely sent"""

    user: User
    device: Device
    program: ProgramDefinition | None = field(default=None)
    circuit: list[Circuit] | None = None
    qprogram: str | None = None
    anneal_program_args: dict | None = None
    vqa: VQA | None = None
    nshots: int = 10
    job_status: JobStatus = JobStatus.NOT_SENT
    job_result: JobResult | None = None
    name: str = "-"
    summary: str = "-"
    id: int = 0

    def __post_init__(self):
        n = len([arg for arg in [self.qprogram, self.circuit, self.anneal_program_args, self.vqa] if arg is not None])
        match n:
            case n if n > 1:
                raise ValueError(
                    "VQA, circuit, qprogram and anneal_program_args were provided, but execute() only takes one of them."
                )
            case 0:
                raise ValueError("Neither of circuit, vqa, qprogram or anneal_program_args were provided.")

    @property
    def user_id(self) -> int | None:
        """User identifier

        Returns:
            int: User identifier
        """
        return self.user.user_id

    @property
    def device_id(self) -> int:
        """Device identifier

        Returns:
            int: Device identifier
        """
        return self.device.id

    @property
    def algorithms(self) -> List[dict]:
        """Get all algorithms described


        Returns:
            List[dict]: a list of all algorithms as a dictionary
        """
        if self.program is None:
            raise ValueError("Job does not contain an algorithm Program")

        return [algorithm.__dict__ for algorithm in self.program.algorithms]

    @property
    def job_request(self) -> JobRequest:
        """Returns a Job Request with the Job instance info

        Returns:
            JobRequest: Job Request object
        """
        return JobRequest(
            user_id=self.user.user_id,
            device_id=self.device.id,
            number_shots=self.nshots,
            job_type=self.job_type,
            name=self.name,
            summary=self.summary,
            description=self._get_job_description(),
        )

    @property
    def job_id(self) -> int:
        """Returns Job identifier

        Returns:
            int: Job identifier
        """
        return self.id

    @job_id.setter
    def job_id(self, job_id: int) -> None:
        """
        Modifies internal Job identifier, accessed via job_id property
        Args:
            job_id: new job_id value
        """
        self.id = job_id

    @property
    def job_type(self):
        """Get the type of the job, checking whether the user has defined circuit or experiment."""
        if self.circuit is not None and self.qprogram is None and self.anneal_program_args is None and self.vqa is None:
            return JobType.CIRCUIT
        if self.qprogram is not None and self.circuit is None and self.anneal_program_args is None and self.vqa is None:
            return JobType.QPROGRAM
        if self.anneal_program_args is not None and self.circuit is None and self.qprogram is None and self.vqa is None:
            return JobType.ANNEALING_PROGRAM
        if self.vqa is not None and self.circuit is None and self.qprogram is None and self.anneal_program_args is None:
            return JobType.VQA
        raise ValueError("Could not determine JobType")

    def _get_job_description(self) -> str:
        """Serialize either circuit or qprogram to obtain job description"""

        if self.qprogram is not None:
            return json.dumps(compress_any(self.qprogram))
        if self.anneal_program_args is not None:
            return json.dumps(compress_any(self.anneal_program_args))
        if self.vqa is not None:
            vqa_as_dict = asdict(self.vqa)
            vqa_as_dict.pop("vqa_dict")
            return json.dumps({**compress_any(self.vqa.vqa_dict), **vqa_as_dict})
        if self.circuit is not None:
            return json.dumps(compress_any([c.to_qasm() for c in self.circuit]))

        raise ValueError("No suitable information found for building description.")

    @property
    def result(self) -> Any:
        """Return the result of the job

        Raises:
            ValueError: Job result still not completed

        Returns:
            Any: Job Result data
        """
        if self.job_result is None:
            raise ValueError("Job result still not completed")
        return self.job_result.data

    @typechecked
    def update_with_job_response(self, job_response: JobResponse) -> None:
        """Updates the current job with the given job response

        Args:
            job_response (JobResponse): Job response

        Raises:
            ValueError: Job response does not belong to the user.
            ValueError: Job response does not belong to the device.
        """
        if self.user.user_id != job_response.user_id:
            raise ValueError("Job response does not belong to the user.")
        if self.device.id != job_response.device_id:
            raise ValueError("Job response does not belong to the device.")
        self.job_id = job_response.job_id
        self.job_status = (
            job_response.status if isinstance(job_response.status, JobStatus) else JobStatus(job_response.status)
        )
        self.job_result = JobResult(job_id=self.id, job_type=job_response.job_type, http_response=job_response.result)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""InitialValue enum"""

from .str_enum import StrEnum


class InitialValue(StrEnum):
    """Initial Value

    Args:
        enum (str):name types of the initial values.
        Supported values are: 'zero', 'one', and 'random'
    """

    ZERO = "zero"
    ONE = "one"
    RANDOM = "random"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Util Functions used by the API module"""

import json
from typing import Any, List

from qibo.models.circuit import Circuit  # type: ignore[import-untyped]

from qiboconnection.config import logger
from qiboconnection.models import JobResult
from qiboconnection.typings.enums import JobStatus, JobType
from qiboconnection.typings.responses.job_response import JobResponse
from qiboconnection.util import decompress_any


def parse_job_responses_to_results(job_responses: List[JobResponse]) -> List[dict | Any | None]:
    """Parse a list of job_responses to a list of dict with the content of each job. If the job is not COMPLETED,
    put a None in its place. For this, we build a JobResult instance for each COMPLETED job, and then we keep its
    `.data`.

    Args:
        job_responses: list of JobResponse instances from which we'll

    Returns:

    """
    raw_results = [parse_job_response_to_result(job_response=job_response) for job_response in job_responses]
    return list(raw_results)


def parse_job_response_to_result(job_response: JobResponse):
    """Parse a single job_response to a single dict with the content of a job. If the job is not COMPLETED,
    put a None in its place. For this, we build a JobResult instance for each COMPLETED job, and then we keep its
    `.data`.

    Args:
        job_response: JobResponse instance from which we'll get the results

    Returns:

    """
    return (
        JobResult(job_id=job_response.job_id, job_type=job_response.job_type, http_response=job_response.result).data
        if job_response.status == JobStatus.COMPLETED
        else None
    )


def deserialize_job_description(raw_description: str, job_type: str) -> dict:
    """Convert base64 job description to its corresponding Qibo Circuit or Qililab experiment

    Args:
        raw_description (str):
        job_type (str):

    Raises:
        ValueError: Job type isn't nor Qibo Circuit neither Qililab experiment

    Returns:
        Circuit | dict: _description_
    """

    description_dict = json.loads(raw_description)
    decompressed_data = decompress_any(**description_dict)
    compressed_data = description_dict.pop("data")
    if job_type == JobType.CIRCUIT:
        return {
            **description_dict,
            "data": [Circuit.from_qasm(decom_data) for decom_data in decompressed_data],
        }
    if job_type == JobType.VQA:
        return {**description_dict, "vqa_dict": decompressed_data}
    if job_type in {JobType.QPROGRAM, JobType.ANNEALING_PROGRAM, JobType.OTHER}:
        return {**description_dict, "data": decompressed_data}
    return {**description_dict, "data": compressed_data}


def log_job_status_info(job_response: JobResponse):
    """Logs a message depending on the `JobResponse.status` belonging to a certain job.

    Args:
        job_response:

    Returns:

    """
    if job_response.status in {JobStatus.QUEUED, JobStatus.PENDING}:
        logger.warning(
            "Your job with id %i is still pending. Job queue position: %s",
            job_response.job_id,
            job_response.queue_position,
        )
        return

    if job_response.status == JobStatus.RUNNING:
        logger.warning("Your job with id %i is still running.", job_response.job_id)
        return
    if job_response.status == JobStatus.NOT_SENT:
        logger.warning("Your job with id %i has not been sent.", job_response.job_id)
        return
    if job_response.status == JobStatus.ERROR:
        logger.error("Your job with id %i failed.", job_response.job_id)
        return
    if job_response.status == JobStatus.COMPLETED:
        logger.warning("Your job with id %i is completed.", job_response.job_id)
        return

    logger.warning(f"Your job with id %i is {job_response.status}.", job_response.job_id)
    return



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Installation script for python
import os
import re
from pathlib import Path

from setuptools import find_packages, setup

PACKAGE = "qiboconnection"


# Returns the library version
def get_version():
    """Gets the version from the package's __init__ file
    if there is some problem, let it happily fail"""
    VERSIONFILE = os.path.join("src", PACKAGE, "__init__.py")
    with open(VERSIONFILE, "rt", encoding="utf-8") as version_file:
        initfile_lines = version_file.readlines()
        VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
        for line in initfile_lines:
            mo = re.search(VSRE, line, re.MULTILINE)
            if mo:
                return mo.group(1)
        return None


# Read in requirements
with open("requirements.txt", encoding="utf-8") as requirements_file:
    requirements = [r.strip() for r in requirements_file]


# load long description from README
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")


setup(
    name="qiboconnection",
    version=get_version(),
    description="Python interface to Qilimanjaro's Services for quantum job executions.",
    author="Qilimanjaro",
    author_email="info@qilimanjaro.tech",
    url="https://github.com/qilimanjaro-tech/qiboconnection",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"": ["*.out"]},
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    install_requires=requirements,
    extras_require={
        "docs": [
            "sphinx",
            "sphinx_rtd_theme",
            "recommonmark",
            "sphinxcontrib-bibtex",
            "sphinx_markdown_tables",
            "nbsphinx",
            "IPython",
        ],
        "tests": ["pytest"],
    },
    python_requires=">=3.10.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass


@dataclass
class VQA:
    """
    Class to be filled for executing variational quantum algorithms through Qilimanjaro's QaaS.
    """

    vqa_dict: dict
    init_params: list
    optimizer_params: dict | None = None

    def __post_init__(self):
        if self.optimizer_params is None:
            self.optimizer_params = {}
